#define UNICODE
#define VER_PRODUCTBUILD 3790

#include "ntoskrnl.h"
#include "rtl.h"

#pragma warning(disable:4100)
#pragma warning(disable:4311)
#pragma warning(disable:4312)

#define IOPORT_READ_KERNEL_SDT			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0001, METHOD_NEITHER, FILE_ANY_ACCESS)
#define IOPORT_WRITE_KERNEL_SDT			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0002, METHOD_NEITHER, FILE_ANY_ACCESS)
#define IOPORT_GET_KERNEL_SDT_INFO		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0003, METHOD_NEITHER, FILE_ANY_ACCESS)
#define IOPORT_GET_KERNEL_SYSTEM_INFO	CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0004, METHOD_NEITHER, FILE_ANY_ACCESS)
#define IOPORT_GET_SYSENTER_INFO		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0005, METHOD_NEITHER, FILE_ANY_ACCESS)
#define IOPORT_LIST_CID					CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0006, METHOD_NEITHER, FILE_ANY_ACCESS)
#define IOPORT_KILL_PROCESS				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0007, METHOD_NEITHER, FILE_ANY_ACCESS)

#define PAGE_SIZE 0x1000
#define STATUS_INFO_LENGTH_MISMATCH      ((NTSTATUS)0xC0000004L)

typedef PVOID *PPVOID;

typedef struct _KSERVICE_PTR {
		DWORD	dwServiceNumber;
        PVOID	dwServiceEntryPoint;
} KSERVICE_PTR, *PKSERVICE_PTR;

typedef struct _EPROCESSINFO_ENTRY {
	PVOID	EPROCESSptr;
	ULONG	UniqueProcessId;
	ULONG	ObjectTable;
	UCHAR	ImageFileName[16];
} EPROCESSINFO_ENTRY, *PEPROCESSINFO_ENTRY;

typedef struct _EPROCESSINFO {
	ULONG				CountOfEntries;
	EPROCESSINFO_ENTRY	ProcessList[1];
} EPROCESSINFO, *PEPROCESSINFO;

typedef struct _XP_HANDLE_TABLE
{
        ULONG                    TableCode;
        PEPROCESS                QuotaProcess;
        PVOID                    UniqueProcessId;
        EX_PUSH_LOCK             HandleTableLock[4];
        LIST_ENTRY               HandleTableList;
        EX_PUSH_LOCK             HandleContentionEvent;
		/*PHANDLE_TRACE_DEBUG_INFO*/PVOID DebugInfo;
        LONG                     ExtraInfoPages;
        ULONG                    FirstFree;
        ULONG                    LastFree;
        ULONG                    NextHandleNeedingPool;
        LONG                     HandleCount;
        LONG                     Flags;
        UCHAR                    StrictFIFO;
} XP_HANDLE_TABLE, *PXP_HANDLE_TABLE;

typedef struct _WIN2K_HANDLE_TABLE
{
		ULONG					Flags;
		LONG					HandleCount;
		PHANDLE_TABLE_ENTRY		**Table;
		PEPROCESS				QuotaProcess;
		HANDLE					UniqueProcessId;
        LONG					FirstFreeTableEntry;
		LONG					NextIndexNeedingPool;
        ERESOURCE				HandleTableLock;
        LIST_ENTRY				HandleTableList;
		KEVENT					HandleContentionEvent;
} WIN2K_HANDLE_TABLE , *PWIN2K_HANDLE_TABLE;

#define WIN2K_TABLE_ENTRY_LOCK_BIT    0x80000000
#define TABLE_LEVEL_MASK              3
#define XP_TABLE_ENTRY_LOCK_BIT       1

#define OBJECT_TO_OBJECT_HEADER(o) CONTAINING_RECORD((o), OBJECT_HEADER, Body);

UNICODE_STRING DevNameStr, DevLinkStr;
PXP_HANDLE_TABLE PspCidTable;
//offsets in ETHREAD
ULONG ThreadsProcessOffset;
// offsets in EPROCESS
ULONG ObjectTableOffset, ImageFileNameOffset, UniqueProcessIdOffset;

NTSTATUS __stdcall HandleOpenClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	IofCompleteRequest(Irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

PVOID __stdcall AllocateInfoBuffer(IN SYSTEM_INFORMATION_CLASS InfoClass, PULONG ReturnLength)
{
	ULONG uSize = PAGE_SIZE;
	PVOID pBuffer;
	NTSTATUS Status;

	do
	{
		pBuffer = ExAllocatePoolWithTag(PagedPool, uSize, 0);
		if (pBuffer != NULL) 
		{
			Status = ZwQuerySystemInformation(InfoClass, pBuffer, uSize, NULL);
		} else return NULL;
		if (Status == STATUS_INFO_LENGTH_MISMATCH)
		{
			ExFreePool(pBuffer);
			uSize *= 2;
		}
	} while (Status == STATUS_INFO_LENGTH_MISMATCH);
	if (Status == STATUS_SUCCESS)
	{
		if (ReturnLength != NULL) *ReturnLength = uSize;
		return pBuffer;
	}
	ExFreePool(pBuffer);
	return NULL;
}

void ProcessObject(PHANDLE_TABLE_ENTRY Entry, PEPROCESSINFO outbuf)
{
	PVOID obj;
	if ( NtBuildNumber == 2195)
	{
		obj = (PVOID)((ULONG)Entry->Object | WIN2K_TABLE_ENTRY_LOCK_BIT);
	} else
	{
		obj = (PVOID)((ULONG)Entry->Object & ~XP_TABLE_ENTRY_LOCK_BIT);
	}

	POBJECT_HEADER ObjectHeader = (POBJECT_HEADER)((PBYTE)obj - 0x18);
	PEPROCESS ep1;
	PETHREAD et1;

	UINT c = outbuf->CountOfEntries;
	
	if (ObjectHeader->ObjectType == (PVOID)PsThreadType)
	{
		et1 = (PETHREAD)obj;
		ep1 = IoThreadToProcess(et1);
		
		if ( NtBuildNumber == 2195)
		{
			outbuf->ProcessList[c].ObjectTable = 1;
		} else
		{
			outbuf->ProcessList[c].ObjectTable = *(PULONG)((PBYTE)ep1 + ObjectTableOffset);
		}

		outbuf->ProcessList[c].EPROCESSptr = ep1;
		outbuf->ProcessList[c].UniqueProcessId = *(PULONG)((PBYTE)ep1 + UniqueProcessIdOffset);
		strcpyA((char *)&outbuf->ProcessList[c].ImageFileName, (char *)((PBYTE)ep1 + ImageFileNameOffset));
		outbuf->CountOfEntries++;
	}
}


void ScanXpHandleTable(PEPROCESSINFO outbuf)
{
	int i, j, k;
	PHANDLE_TABLE_ENTRY Entry;
	ULONG TableCode = PspCidTable->TableCode & ~TABLE_LEVEL_MASK;

	outbuf->CountOfEntries = 0;

	switch (PspCidTable->TableCode & TABLE_LEVEL_MASK)
	{
		case 0:
			for (i = 0; i < 0x200; i++)
			{
				Entry = &((PHANDLE_TABLE_ENTRY)TableCode)[i];
				if (Entry->Object) ProcessObject(Entry, outbuf);
			};
			break;

		case 1:
			for (i = 0; i < 0x200; i++)
			{
				if (((PVOID *)TableCode)[i])
				{
					for (j = 0; j < 0x200; j++)
					{
						Entry = &((PHANDLE_TABLE_ENTRY *)TableCode)[i][j];
						if (Entry->Object) ProcessObject(Entry, outbuf);
					}
				}
			}
			break;
		case 2:
			for (i = 0; i < 0x200; i++)
			{
				if (((PVOID *)TableCode)[i])
				{
					for (j = 0; j < 0x200; j++)
					{
						if (((PVOID **)TableCode)[i][j])
						{
							for (k = 0; k < 0x200; k++)
							{
								Entry = &((PHANDLE_TABLE_ENTRY **)TableCode)[i][j][k];
								if (Entry->Object) ProcessObject(Entry, outbuf);
							}
						}
					}
				}
			}
			break;
	}
}

void ScanWin2KHandleTable(PEPROCESSINFO outbuf)
{
	int i, j, k;
	PHANDLE_TABLE_ENTRY Entry;
	PWIN2K_HANDLE_TABLE HandleTable = (PWIN2K_HANDLE_TABLE)PspCidTable;

	outbuf->CountOfEntries = 0;

	for (i = 0; i < 0x100; i++)
	{
		if (HandleTable->Table[i])
		{
			for (j = 0; j < 0x100; j++)
			{
				if (HandleTable->Table[i][j])
				{
					for (k = 0; k < 0x100; k++)
					{
						Entry = &HandleTable->Table[i][j][k];

						if (Entry->Object) ProcessObject(Entry, outbuf);
					}
				}
			}
		}
	}
}

typedef NTSTATUS (__stdcall *PRTLGETVERSION) (IN OUT PRTL_OSVERSIONINFOW  lpVersionInformation);

bool InitPointers()
{
	UNICODE_STRING			fname;
	RTL_OSVERSIONINFOEXW	inf1;
	PRTLGETVERSION			RtlGetVersion;

	PBYTE ptmp = (PBYTE)&PsLookupProcessByProcessId;

	for (int c = 0; c < 256; c++)
	{
		if ((ptmp[c] == 0xff) && (ptmp[c + 1] == 0x35) && (ptmp[c + 6] == 0xe8))
		{
			ptmp += c + 2;
			break;
		}
	}

	inf1.dwOSVersionInfoSize = sizeof(RTL_OSVERSIONINFOEXW);
	inf1.wServicePackMajor = 0;
	RtlInitUnicodeString(&fname, L"RtlGetVersion");
	RtlGetVersion = (PRTLGETVERSION)MmGetSystemRoutineAddress(&fname);
	if (RtlGetVersion != NULL) RtlGetVersion((PRTL_OSVERSIONINFOW)&inf1);

	switch ( NtBuildNumber )
	{
	case 3790:
		{
			switch ( inf1.wServicePackMajor )
			{
			case 0:
				{
					//DbgPrint("Microsoft Windows Server 2003"); // no SP
					ObjectTableOffset = 0xc4;
					ImageFileNameOffset = 0x154;
					UniqueProcessIdOffset = 0x84;
					break;
				}
			case 1:
				{
					//DbgPrint("Microsoft Windows Server 2003 Service Pack 1");
					ObjectTableOffset = 0xd4;
					ImageFileNameOffset = 0x164;
					UniqueProcessIdOffset = 0x94;
					break;
				}
			default:
				//DbgPrint("Microsoft Windows Server 2003 UNKNOWN SERVICE PACK");
				return false;
				break;
			}
		}
	case 2600:
		{
			//DbgPrint("Microsoft Windows XP"); //SP2
			ObjectTableOffset = 0xc4;
			ImageFileNameOffset = 0x174;
			UniqueProcessIdOffset = 0x84;
			break;
		}
	case 2195:
		{
			//DbgPrint("Microsoft Windows 2000"); //SP4
			ObjectTableOffset = 0x124;
			ImageFileNameOffset = 0x1fc;
			UniqueProcessIdOffset = 0x9c;
			break;
		}
	default:
		{
			//DbgPrint("UNSUPPORTED OS");
			return false;
			break;
		}
	}
	
	PspCidTable = (PXP_HANDLE_TABLE)**(PVOID **)ptmp;

	return true;
}

NTSTATUS __stdcall HandleIOCTL(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	NTSTATUS status = STATUS_SUCCESS;

	switch (Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.IoControlCode)
	{
	case IOPORT_GET_KERNEL_SDT_INFO:
		if ((Irp->UserBuffer != NULL) & (Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.OutputBufferLength >= sizeof(SERVICE_DESCRIPTOR_ENTRY)))
		{
			Irp->IoStatus.Information = sizeof(SERVICE_DESCRIPTOR_ENTRY);
			memcopy(Irp->UserBuffer, &KeServiceDescriptorTable, sizeof(SERVICE_DESCRIPTOR_ENTRY));
		}
		break;

	case IOPORT_READ_KERNEL_SDT:
		DWORD sz;
  		
		sz = KeServiceDescriptorTable.NumberOfServices * sizeof(PVOID);
		if ((Irp->UserBuffer != NULL) & (Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.OutputBufferLength >= sz))
		{
			Irp->IoStatus.Information = sz;
			memcopy(Irp->UserBuffer, KeServiceDescriptorTable.ServiceTableBase, sz);
		}
		break;

	case IOPORT_WRITE_KERNEL_SDT:
		PKSERVICE_PTR sp1;	
		sp1 = (PKSERVICE_PTR)Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.Type3InputBuffer;
		if ((sp1 != NULL) & (sp1->dwServiceNumber < KeServiceDescriptorTable.NumberOfServices))
		{
			KeServiceDescriptorTable.ServiceTableBase[sp1->dwServiceNumber] = (sp1->dwServiceEntryPoint);
		}
		break;
	case IOPORT_GET_KERNEL_SYSTEM_INFO:
		
		ULONG outsz, bytesIO;
		PVOID pRetBuf;
		outsz = Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.OutputBufferLength;
		if (Irp->UserBuffer != NULL)
		{
			SYSTEM_INFORMATION_CLASS InfoClass = SystemBasicInformation;			
			if (Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.Type3InputBuffer != NULL)
			{
				InfoClass = *(SYSTEM_INFORMATION_CLASS *)Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.Type3InputBuffer;
			} 

			pRetBuf = AllocateInfoBuffer(InfoClass, &bytesIO);
			if (pRetBuf != NULL)
			{
				if (outsz >= bytesIO)
				{
					memcopy(Irp->UserBuffer, pRetBuf, bytesIO);
					Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = bytesIO;
				} else {
					Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = sizeof(ULONG);
					*(PULONG)Irp->UserBuffer = bytesIO;
				}
				ExFreePool(pRetBuf);
			}
		}
		break;
	 case IOPORT_GET_SYSENTER_INFO:
         DWORD eiplow;
         if ((Irp->UserBuffer != NULL) & 
(Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.OutputBufferLength >= sizeof(DWORD)))
         {
             __asm
             {
                   mov ecx, 0x0000176
                   rdmsr
                   mov eiplow, eax
			 }
			 Irp->IoStatus.Status = STATUS_SUCCESS;
             Irp->IoStatus.Information = sizeof(DWORD);
             *(PDWORD)Irp->UserBuffer = eiplow;
          }
          break;
	case IOPORT_LIST_CID:
		{
			if (Irp->UserBuffer != NULL)
			{
				PEPROCESSINFO pinf = (PEPROCESSINFO)Irp->UserBuffer;
	
				if ( NtBuildNumber == 2195)
				{
					ScanWin2KHandleTable(pinf);
				} else
				{
					ScanXpHandleTable(pinf);
				}
				Irp->IoStatus.Information = sizeof(ULONG) + (sizeof(EPROCESSINFO_ENTRY) * pinf->CountOfEntries);
			}
			break;
		}
	case IOPORT_KILL_PROCESS:
		{
			if (Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.Type3InputBuffer != NULL)
			{
				ULONG uProcessId = *(PULONG)Irp->Tail.Overlay.CurrentStackLocation->Parameters.DeviceIoControl.Type3InputBuffer;
				
				CLIENT_ID cid1 =
				{
					uProcessId, 
					0
				};
				OBJECT_ATTRIBUTES attr =  
				{ 
					sizeof OBJECT_ATTRIBUTES, 
						0, 
						NULL, 
						0,
						NULL,
						NULL
				};
				HANDLE hProcess = 0;
				status = ZwOpenProcess(&hProcess, PROCESS_TERMINATE, &attr, &cid1);
				if (status == STATUS_SUCCESS)
				{
					status = ZwTerminateProcess(hProcess, 0xBADDEAD); 
					ZwClose(hProcess);
				}
				Irp->IoStatus.Status = status;
			} 
			break;
		}
	default:
		status = STATUS_NOT_IMPLEMENTED;
	};
	IofCompleteRequest(Irp, IO_NO_INCREMENT);
	return status;
}


VOID __stdcall Unload(IN PDRIVER_OBJECT  DriverObject)
{
	IoDeleteSymbolicLink(&DevLinkStr);
	IoDeleteDevice(DriverObject->DeviceObject);
}

NTSTATUS __stdcall DriverEntry(IN PDRIVER_OBJECT  DriverObject, IN PUNICODE_STRING  RegistryPath)
{
	PDEVICE_OBJECT pdev1;
	NTSTATUS ntst;

	if (!InitPointers()) return 0xC0000183;

	RtlInitUnicodeString(&DevNameStr, L"\\Device\\rkhdrv10");

	ntst = IoCreateDevice(DriverObject, 0, &DevNameStr, FILE_DEVICE_UNKNOWN, 0, FALSE, &pdev1);

	if (ntst == STATUS_SUCCESS)
	{
		RtlInitUnicodeString(&DevLinkStr, L"\\DosDevices\\rkhdrv10");
		if (IoCreateSymbolicLink(&DevLinkStr, &DevNameStr) != STATUS_SUCCESS)
		{
			IoDeleteDevice(DriverObject->DeviceObject);
			return STATUS_OBJECT_NAME_EXISTS;
		}
		DriverObject->DriverUnload = &Unload;
		DriverObject->MajorFunction[IRP_MJ_CREATE] = &HandleOpenClose;
		DriverObject->MajorFunction[IRP_MJ_CLOSE] = &HandleOpenClose;
		DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = &HandleIOCTL;
	}

	return ntst;
}